@@ -50,6 +50,26 @@ class DBAdapter(object):
         self.engine_opts = engine_opts or {}
         self.session_opts = session_opts or {}
 
+    @staticmethod
+    def from_url(
+        driver: str,
+        url: str | urllib.parse.ParseResult,
+        engine_opts: dict[str, Any] | None = None,
+        session_opts: dict[str, Any] | None = None,
+    ) -> "DBAdapter":
+        if isinstance(url, str):
+            return DBAdapter.from_url(driver, urllib.parse.urlparse(url), engine_opts, session_opts)
+        if isinstance(url, urllib.parse.ParseResult):
+            return DBAdapter(driver,
+                             url.hostname,
+                             url.port,
+                             url.username,
+                             url.password,
+                             url.path.strip("/"),
+                             engine_opts,
+                             session_opts)
+        raise ValueError("illegal parameter 'url'")
+
     @property
     def connection_string(self) -> str:
         port_part = "" if self.port is None else (":%d" % self.port)
