@@ -1,14 +1,12 @@
-import random
 import unittest
 
 import ddt
 
-from iker.common.utils.sequtils import IntervalRelation
 from iker.common.utils.sequtils import batch_yield
 from iker.common.utils.sequtils import chunk, chunk_between, chunk_with_key, merge_chunks
 from iker.common.utils.sequtils import deduped, grouped
 from iker.common.utils.sequtils import head, init, last, tail
-from iker.common.utils.sequtils import interval_relation, intervals_intersect, intervals_subtract, intervals_union
+from iker.common.utils.sequtils import seq
 
 
 @ddt.ddt
@@ -792,288 +790,168 @@ class SeqUtilsTest(unittest.TestCase):
         self.assertEqual(expect, merge_chunks(data, merge_func=merge_func, drop_exclusive_end=True))
 
     @ddt.data(
-        ((-2, 2), (-5, -3), IntervalRelation.LeftDetach),
-        ((-2, 2), (-5, -2), IntervalRelation.LeftTouch),
-        ((-2, 2), (-5, 0), IntervalRelation.LeftOverlap),
-        ((-2, 2), (-2, 0), IntervalRelation.LeftAlignOverlay),
-        ((-2, 2), (-2, 3), IntervalRelation.LeftAlignCover),
-        ((-2, 2), (-1, 1), IntervalRelation.Overlay),
-        ((-2, 2), (-2, 2), IntervalRelation.Identical),
-        ((-2, 2), (-3, 3), IntervalRelation.Cover),
-        ((-2, 2), (-3, 2), IntervalRelation.RightAlignCover),
-        ((-2, 2), (0, 2), IntervalRelation.RightAlignOverlay),
-        ((-2, 2), (0, 5), IntervalRelation.RightOverlap),
-        ((-2, 2), (2, 5), IntervalRelation.RightTouch),
-        ((-2, 2), (3, 5), IntervalRelation.RightDetach),
-        ((-2, 2), (-5, -5), IntervalRelation.LeftDetach),
-        ((-2, 2), (-2, -2), IntervalRelation.LeftOn),
-        ((-2, 2), (0, 0), IntervalRelation.Overlay),
-        ((-2, 2), (2, 2), IntervalRelation.RightOn),
-        ((-2, 2), (5, 5), IntervalRelation.RightDetach),
-        ((0, 0), (-5, -5), IntervalRelation.LeftDetach),
-        ((0, 0), (0, 0), IntervalRelation.Identical),
-        ((0, 0), (5, 5), IntervalRelation.RightDetach),
-        ((0, 0), (-5, -2), IntervalRelation.LeftDetach),
-        ((0, 0), (-5, 0), IntervalRelation.RightAlignCover),
-        ((0, 0), (-2, 2), IntervalRelation.Cover),
-        ((0, 0), (0, 5), IntervalRelation.LeftAlignCover),
-        ((0, 0), (2, 5), IntervalRelation.RightDetach),
+        ([0], 0, [0]),
+        ([1, 2, 3, 4, 5], 0, [1]),
+        ([1, 2, 3, 4, 5], -1, [5]),
+        ([1, 2, 3, 4, 5], slice(None, 1), [1]),
+        ([1, 2, 3, 4, 5], slice(1, None), [2, 3, 4, 5]),
+        ([1, 2, 3, 4, 5], slice(None, -1), [1, 2, 3, 4]),
+        ([1, 2, 3, 4, 5], slice(-1, None), [5]),
+        ([1, 2, 3, 4, 5], slice(1, -1), [2, 3, 4]),
+        ([1, 2, 3, 4, 5], slice(1, -1, 2), [2, 4]),
     )
     @ddt.unpack
-    def test_interval_relation(self, b, a, expect):
-        self.assertEqual(expect, interval_relation(a, b))
+    def test_seq_builtin_getitem(self, data, item, expect):
+        actual = seq(data)[item]
+        self.assertEqual(actual.data, expect)
 
     @ddt.data(
-        ([], [], []),
-        ([(-2, 2)], [], [(-2, 2)]),
-        ([(-2, 2)], [(3, 5)], [(-2, 2), (3, 5)]),
-        ([(-2, 2)], [(-5, -3)], [(-5, -3), (-2, 2)]),
-        ([(-2, 2)], [(2, 5)], [(-2, 5)]),
-        ([(-2, 2)], [(-5, -2)], [(-5, 2)]),
-        ([(-2, 2)], [(0, 4)], [(-2, 4)]),
-        ([(-2, 2)], [(-4, 0)], [(-4, 2)]),
-        ([(-2, 2)], [(-4, 4)], [(-4, 4)]),
-        ([(-2, 2)], [(-2, 2)], [(-2, 2)]),
-        ([(-2, 2)], [(-1, 1)], [(-2, 2)]),
-        ([(-2, 2)], [(0, 2)], [(-2, 2)]),
-        ([(-2, 2)], [(-2, 0)], [(-2, 2)]),
-        ([(-2, 2)], [(2, 2)], [(-2, 2)]),
-        ([(-2, 2)], [(-2, -2)], [(-2, 2)]),
-        ([(-2, 2)], [(0, 0)], [(-2, 2)]),
-        ([(-10, 10)], [(-8, -4), (-2, 2), (4, 8)], [(-10, 10)]),
-        ([(-10, 10)], [(-10, -6), (-2, 2), (6, 10)], [(-10, 10)]),
-        ([(-10, 10)], [(-12, -8), (-2, 2), (8, 12)], [(-12, 12)]),
-        ([(-10, 10)], [(-12, -10), (-2, 2), (10, 12)], [(-12, 12)]),
-        ([(-10, 10)], [(-16, -12), (-2, 2), (12, 16)], [(-16, -12), (-10, 10), (12, 16)]),
-        ([(-10, 10)], [(-16, -12), (-11, 11), (12, 16)], [(-16, -12), (-11, 11), (12, 16)]),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-5, -3), (3, 5)],
-            [(-10, -6), (-5, -3), (-2, 2), (3, 5), (6, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-6, -2), (2, 6)],
-            [(-10, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, -1), (1, 8)],
-            [(-10, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 8)],
-            [(-10, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(0, 8)],
-            [(-10, -6), (-2, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 0)],
-            [(-10, 2), (6, 10)],
-        ),
+        ([], 0, lambda x, y: x, []),
+        ([], None, lambda x, y: x, []),
+        ([0], 1, lambda x, y: x + y, [1]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x + y, [2, 4, 7, 11, 16]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x * y, [1, 2, 6, 24, 120]),
+        ([1, 2, 3, 4, 5], 0, lambda x, y: x * 10 + y, [1, 12, 123, 1234, 12345]),
     )
     @ddt.unpack
-    def test_intervals_union(self, xs, ys, expect):
-        self.assertEqual(expect, intervals_union(xs, ys))
-        self.assertEqual(expect, intervals_union(ys, xs))
-        self.assertEqual(expect, intervals_union(xs, *[[y] for y in ys]))
-        self.assertEqual(expect, intervals_union(ys, *[[x] for x in xs]))
+    def test_seq_scan_left(self, data, zero, func, expect):
+        actual = seq(data).scan_left(zero, func)
+        self.assertEqual(actual.data, expect)
 
     @ddt.data(
-        ([], [], []),
-        ([(-2, 2)], [], []),
-        ([(-2, 2)], [(3, 5)], []),
-        ([(-2, 2)], [(-5, -3)], []),
-        ([(-2, 2)], [(2, 5)], [(2, 2)]),
-        ([(-2, 2)], [(-5, -2)], [(-2, -2)]),
-        ([(-2, 2)], [(0, 4)], [(0, 2)]),
-        ([(-2, 2)], [(-4, 0)], [(-2, 0)]),
-        ([(-2, 2)], [(-4, 4)], [(-2, 2)]),
-        ([(-2, 2)], [(-2, 2)], [(-2, 2)]),
-        ([(-2, 2)], [(-1, 1)], [(-1, 1)]),
-        ([(-2, 2)], [(0, 2)], [(0, 2)]),
-        ([(-2, 2)], [(-2, 0)], [(-2, 0)]),
-        ([(-2, 2)], [(2, 2)], [(2, 2)]),
-        ([(-2, 2)], [(-2, -2)], [(-2, -2)]),
-        ([(-2, 2)], [(0, 0)], [(0, 0)]),
-        ([(-10, 10)], [(-8, -4), (-2, 2), (4, 8)], [(-8, -4), (-2, 2), (4, 8)]),
-        ([(-10, 10)], [(-10, -6), (-2, 2), (6, 10)], [(-10, -6), (-2, 2), (6, 10)]),
-        ([(-10, 10)], [(-12, -8), (-2, 2), (8, 12)], [(-10, -8), (-2, 2), (8, 10)]),
-        ([(-10, 10)], [(-12, -10), (-2, 2), (10, 12)], [(-10, -10), (-2, 2), (10, 10)]),
-        ([(-10, 10)], [(-16, -12), (-2, 2), (12, 16)], [(-2, 2)]),
-        ([(-10, 10)], [(-16, -12), (-11, 11), (12, 16)], [(-10, 10)]),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-5, -3), (3, 5)],
-            [],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-6, -2), (2, 6)],
-            [(-6, -6), (-2, -2), (2, 2), (6, 6)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, -1), (1, 8)],
-            [(-8, -6), (-2, -1), (1, 2), (6, 8)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 8)],
-            [(-8, -6), (-2, 2), (6, 8)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(0, 8)],
-            [(0, 2), (6, 8)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 0)],
-            [(-8, -6), (-2, 0)],
-        ),
+        ([], 0, lambda x, y: x, []),
+        ([], None, lambda x, y: x, []),
+        ([0], 1, lambda x, y: x + y, [1]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x + y, [16, 15, 13, 10, 6]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x * y, [120, 120, 60, 20, 5]),
+        ([1, 2, 3, 4, 5], 0, lambda x, y: x * 10 + y, [54321, 5432, 543, 54, 5]),
     )
     @ddt.unpack
-    def test_intervals_intersect(self, xs, ys, expect):
-        self.assertEqual(expect, intervals_intersect(xs, ys))
-        self.assertEqual(expect, intervals_intersect(ys, xs))
-        self.assertEqual(expect,
-                         intervals_union(intervals_intersect(xs, []), *[intervals_intersect(xs, [y]) for y in ys]))
-        self.assertEqual(expect,
-                         intervals_union(intervals_intersect(ys, []), *[intervals_intersect(ys, [x]) for x in xs]))
+    def test_seq_scan_right(self, data, zero, func, expect):
+        actual = seq(data).scan_right(zero, func)
+        self.assertEqual(actual.data, expect)
 
     @ddt.data(
-        ([], [], []),
-        ([(-2, 2)], [], [(-2, 2)]),
-        ([(-2, 2)], [(3, 5)], [(-2, 2)]),
-        ([(-2, 2)], [(-5, -3)], [(-2, 2)]),
-        ([(-2, 2)], [(2, 5)], [(-2, 2)]),
-        ([(-2, 2)], [(-5, -2)], [(-2, 2)]),
-        ([(-2, 2)], [(0, 4)], [(-2, 0)]),
-        ([(-2, 2)], [(-4, 0)], [(0, 2)]),
-        ([(-2, 2)], [(-4, 4)], []),
-        ([(-2, 2)], [(-2, 2)], []),
-        ([(-2, 2)], [(-1, 1)], [(-2, -1), (1, 2)]),
-        ([(-2, 2)], [(0, 2)], [(-2, 0)]),
-        ([(-2, 2)], [(-2, 0)], [(0, 2)]),
-        ([(-2, 2)], [(2, 2)], [(-2, 2)]),
-        ([(-2, 2)], [(-2, -2)], [(-2, 2)]),
-        ([(-2, 2)], [(0, 0)], [(-2, 0), (0, 2)]),
-        ([(-10, 10)], [(-8, -4), (-2, 2), (4, 8)], [(-10, -8), (-4, -2), (2, 4), (8, 10)]),
-        ([(-10, 10)], [(-10, -6), (-2, 2), (6, 10)], [(-6, -2), (2, 6)]),
-        ([(-10, 10)], [(-12, -8), (-2, 2), (8, 12)], [(-8, -2), (2, 8)]),
-        ([(-10, 10)], [(-12, -10), (-2, 2), (10, 12)], [(-10, -2), (2, 10)]),
-        ([(-10, 10)], [(-16, -12), (-2, 2), (12, 16)], [(-10, -2), (2, 10)]),
-        ([(-10, 10)], [(-16, -12), (-11, 11), (12, 16)], []),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-5, -3), (3, 5)],
-            [(-10, -6), (-2, 2), (6, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-6, -2), (2, 6)],
-            [(-10, -6), (-2, 2), (6, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, -1), (1, 8)],
-            [(-10, -8), (-1, 1), (8, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 8)],
-            [(-10, -8), (8, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(0, 8)],
-            [(-10, -6), (-2, 0), (8, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 0)],
-            [(-10, -8), (0, 2), (6, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-2, 8)],
-            [(-10, -6), (8, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 2)],
-            [(-10, -8), (6, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-4, 8)],
-            [(-10, -6), (8, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 4)],
-            [(-10, -8), (6, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-6, 8)],
-            [(-10, -6), (8, 10)],
-        ),
-        (
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-8, 6)],
-            [(-10, -8), (6, 10)],
-        ),
-        (
-            [(-16, -12), (-10, -6), (-2, 2), (6, 10), (12, 16)],
-            [(-4, 8)],
-            [(-16, -12), (-10, -6), (8, 10), (12, 16)],
-        ),
-        (
-            [(-16, -12), (-10, -6), (-2, 2), (6, 10), (12, 16)],
-            [(-8, 4)],
-            [(-16, -12), (-10, -8), (6, 10), (12, 16)],
-        ),
-        (
-            [(-16, -12), (12, 16)],
-            [(-10, -6), (-2, 2), (6, 10)],
-            [(-16, -12), (12, 16)],
-        ),
-        (
-            [(-16, -12), (12, 16)],
-            [(-20, -14), (-10, -6), (-2, 2), (6, 10), (14, 20)],
-            [(-14, -12), (12, 14)],
-        ),
+        ([], lambda x: x, []),
+        ([0], lambda x: x, [0]),
+        ([1, 2, 3, 4, 5], lambda x: x * 2, [2, 4, 6, 8, 10]),
+        ([1, 2, 3, 4, 5], lambda x: x ** 2, [1, 4, 9, 16, 25]),
     )
     @ddt.unpack
-    def test_intervals_subtract(self, xs, ys, expect):
-        self.assertEqual(expect, intervals_subtract(xs, ys))
-        self.assertEqual(expect, intervals_subtract(xs, *[[y] for y in ys]))
-        self.assertEqual(expect, intervals_subtract(xs, intervals_intersect(xs, ys)))
-        self.assertEqual(expect, intervals_subtract(xs, *[intervals_intersect(xs, [y]) for y in ys]))
+    def test_seq_map(self, data, func, expect):
+        actual = seq(data).map(func)
+        self.assertEqual(actual.data, expect)
 
-    @ddt.idata([(x,) for x in range(100)])
+    @ddt.data(
+        ([], 0, lambda x, y: x, [0]),
+        ([], None, lambda x, y: x, []),
+        ([0], 1, lambda x, y: x + y, [1]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x + y, [16]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x * y, [120]),
+        ([1, 2, 3, 4, 5], 0, lambda x, y: x * 10 + y, [12345]),
+    )
     @ddt.unpack
-    def test_intervals_operations_equality(self, x):
-        def make_intervals(size, lo, hi):
-            vs = [float(random.randrange(lo, hi))]
-            for _ in range(size * 2):
-                vs.append(last(vs) + float(random.randrange(lo, hi)))
-            return [(vs[2 * i], vs[2 * i + 1]) for i in range(size)]
+    def test_seq_fold_left(self, data, zero, func, expect):
+        actual = seq(data).fold_left(zero, func)
+        self.assertEqual(actual.data, expect)
 
-        xs = make_intervals(200, 2, 10)
-        ys = make_intervals(200, 2, 10)
+    @ddt.data(
+        ([], 0, lambda x, y: x, [0]),
+        ([0], 1, lambda x, y: x + y, [1]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x + y, [16]),
+        ([1, 2, 3, 4, 5], 1, lambda x, y: x * y, [120]),
+        ([1, 2, 3, 4, 5], 0, lambda x, y: x * 10 + y, [54321]),
+    )
+    @ddt.unpack
+    def test_seq_fold_right(self, data, zero, func, expect):
+        actual = seq(data).fold_right(zero, func)
+        self.assertEqual(actual.data, expect)
 
-        self.assertEqual(intervals_union(xs, ys), intervals_union(ys, xs))
-        self.assertEqual(intervals_intersect(xs, ys), intervals_intersect(ys, xs))
-        self.assertEqual(intervals_intersect(xs, ys),
-                         intervals_union(intervals_intersect(xs, []), *[intervals_intersect(xs, [y]) for y in ys]))
-        self.assertEqual([], intervals_subtract(xs, xs))
-        self.assertEqual(intervals_subtract(xs, ys),
-                         intervals_subtract(xs, *[intervals_intersect(xs, [y]) for y in ys]))
+    @ddt.data(
+        ([], lambda x, y: x, []),
+        ([0], lambda x, y: x + y, [0]),
+        ([1, 2, 3, 4, 5], lambda x, y: x + y, [15]),
+        ([1, 2, 3, 4, 5], lambda x, y: y, [5]),
+    )
+    @ddt.unpack
+    def test_seq_reduce(self, data, func, expect):
+        actual = seq(data).reduce(func)
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], None, []),
+        ([0], None, [0]),
+        ([1, 2, 3, 4, 5], None, [5]),
+        ([1, 2, 3, 4, 5], lambda x, y: True, [1]),
+        ([1, 2, 3, 4, 5], lambda x, y: False, [5]),
+    )
+    @ddt.unpack
+    def test_seq_max(self, data, func, expect):
+        actual = seq(data).max(func)
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], None, []),
+        ([0], None, [0]),
+        ([1, 2, 3, 4, 5], None, [1]),
+        ([1, 2, 3, 4, 5], lambda x, y: True, [1]),
+        ([1, 2, 3, 4, 5], lambda x, y: False, [5]),
+    )
+    @ddt.unpack
+    def test_seq_min(self, data, func, expect):
+        actual = seq(data).min(func)
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], lambda x: x % 2, []),
+        ([0], lambda x: x % 2, [(0, [0])]),
+        ([1, 2, 3, 4, 5], lambda x: x % 2, [(0, [2, 4]), (1, [1, 3, 5])]),
+        ([1, 2, 3, 4, 5], lambda x: x * 2, [(2, [1]), (4, [2]), (6, [3]), (8, [4]), (10, [5])]),
+    )
+    @ddt.unpack
+    def test_seq_group(self, data, func, expect):
+        actual = seq(data).group(func)
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], []),
+        ([(0, "")], [0]),
+        ([(1, "a"), (2, "b"), (3, "c"), (4, "d"), (5, "e")], [1, 2, 3, 4, 5]),
+    )
+    @ddt.unpack
+    def test_seq_keys(self, data, expect):
+        actual = seq(data).keys()
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], []),
+        ([(0, "")], [""]),
+        ([(1, "a"), (2, "b"), (3, "c"), (4, "d"), (5, "e")], ["a", "b", "c", "d", "e"]),
+    )
+    @ddt.unpack
+    def test_seq_values(self, data, expect):
+        actual = seq(data).values()
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], []),
+        ([(0, "")], [("", 0)]),
+        ([(1, "a"), (2, "b"), (3, "c"), (4, "d"), (5, "e")], [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]),
+    )
+    @ddt.unpack
+    def test_seq_swap(self, data, expect):
+        actual = seq(data).swap()
+        self.assertEqual(actual.data, expect)
+
+    @ddt.data(
+        ([], []),
+        ([[]], []),
+        ([[0]], [0]),
+        ([[1, 2, 3, 4, 5]], [1, 2, 3, 4, 5]),
+        ([[], [1], [2], [3], [4], [5], []], [1, 2, 3, 4, 5]),
+        ([[], [1, 2, 3, 4, 5], []], [1, 2, 3, 4, 5]),
+    )
+    @ddt.unpack
+    def test_seq_flatten(self, data, expect):
+        actual = seq(data).flatten()
+        self.assertEqual(actual.data, expect)
